from airflow.hooks.base_hook import BaseHook



class CustomHook(BaseHook):

    def __init__(self, *args, **kwargs):
        print('Hello in Init')
        conn = self.get_connection('mssql_default')
        print(f"AIRFLOW_CONN_{conn.conn_id.upper()}='{conn.get_uri()}'")


import logging
import csv
from contextlib import closing
import xlrd
# install modules
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.exceptions import AirflowException
from airflow.hooks.base_hook import BaseHook

class NovigiXlsToCSVOperator(BaseOperator):

    def __init__(self, input_excel: str, output_csv: str, sheet_name: str = None, column_range: range = None, row_range: range = None, headers: list = None, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.input_excel = input_excel
            self.output_csv = output_csv
            self.sheet_name = sheet_name
            self.column_range = column_range
            self.row_range = row_range
            self.headers = headers

    def clean(self, value):
            #use this callback to clean
            return value

    def execute(self, context):
            logging.info("Exporting EXCEL data to csv: {} -> CSV: {}".format(self.input_excel, self.output_csv))

            wb = xlrd.open_workbook(filename = self.input_excel)
            if self.sheet_name is None:
                sheet = wb.sheet_by_index(0)
            else:
                sheet = wb.sheet_by_name(self.sheet_name)

            if self.column_range is None:
                self.column_range = range(sheet.ncols)

            if self.row_range is None:
                self.row_range = range(sheet.nrows)

            with open(self.output_csv, "w") as outfile:
                writer = csv.writer(outfile, quoting=csv.QUOTE_NONNUMERIC)
                if self.headers is not None:
                    writer.writerow(headers)

                for i in self.row_range:
                    row = []
                    for j in self.column_range:
                        cell = sheet.cell_value(i, j)
                        row.append(cell)

                    writer.writerow([self.clean(x) for x in row])

            logging.info("Completed Novigi XLS to CSV operator. CSV: {}".format(self.output_csv))

            return True



from airflow.models.baseoperator import BaseOperator

from airflow.utils.decorators import apply_defaults

from novigi_common.hooks.sample_custom_hook import CustomHook


class CustomFileProcessingOperator(BaseOperator):
    """
        Currently this operator is not doing anything practically.
        We are just taking count of the source file and putting it on XCOM or demo purpose.
        But anyone can add logic to based on hteir reqirement. The configurations are structure will be similar.
    """

    @apply_defaults
    def __init__(
            self,
            name: str,
            *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.name = name

    def execute(self, context):
        message = "Hello {}".format(self.name)
        print(message)
        source_hook = CustomHook()
        return message


from airflow.models.baseoperator import BaseOperator

from airflow.utils.decorators import apply_defaults

from datetime import datetime
import csv
import logging
import pandas as pd
import datetime
import os


class CustomCSVToS3Operator(BaseOperator):

    template_fields = ("query","inputCSVPath")

    @apply_defaults
    def __init__(
        self,
        inputCSVPath: str,
        outputCSVDir: str,
        reportName: str,
        encoding: str = "cp1252",
        filterOutColumns=None,
        query: str = None,
        dateColumns=None,
        *args,
        **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.filterOutColumns = (
            filterOutColumns or []
        )  # columns which are needed to filterout from the source CSV
        self.inputCSVPath = inputCSVPath  # source file path
        self.outputCSVDir = outputCSVDir  # file output path
        self.reportName = reportName
        self.encoding = encoding          # encoding type
        self.csvFileName = reportName + ".csv"
        self.query = query  # query to filterout rows from csv based on conditions. ex : ('Salary_in_1000 >= 100 & Age < 60 ')
        self.dateColumns = dateColumns or []

    def stringToDate(self, inputDate):
        for fmt in ("%d/%m/%Y %H:%M:%S %p", "%d/%m/%Y %H:%M:%S", "%d/%m/%Y"):
            try:
                return datetime.datetime.strptime(inputDate, fmt)
            except:
                continue
        return datetime.datetime(2050, 12, 31)

    def execute(self, context):
        current_execution_date = context["execution_date"].strftime("%Y/%m/%d")
        fullCSVPath = os.path.join(
            self.outputCSVDir, self.reportName, current_execution_date, self.csvFileName
        )

        logging.info("Exporting data from " + self.inputCSVPath)
        data_frame = pd.read_csv(self.inputCSVPath, index_col=False, encoding= self.encoding )

        if len(self.dateColumns) > 0:
            logging.info("Formatting date columns " + str(self.dateColumns))
            for column in self.dateColumns:
                data_frame[column] = data_frame[column].apply(
                    lambda x: self.stringToDate(x)
                )

        if len(self.filterOutColumns) > 0:
            logging.info("Dropping the columns " + str(self.filterOutColumns))
            data_frame = data_frame.drop(self.filterOutColumns, axis=1)

        if self.query.strip():
            logging.info("Applying the query " + self.query)
            data_frame = data_frame.query(self.query)

        logging.info("Writing data into the " + fullCSVPath)
        data_frame.to_csv(fullCSVPath, index=False)

        logging.info("Copying CSV to S3 completed ")
        return True



from typing import List, Optional, Union

from airflow.models.baseoperator import BaseOperator

from airflow.utils.decorators import apply_defaults

#from airflow.providers.postgres.hooks.postgres import PostgresHook
from novigi_common.hooks.postgres_hook import PostgresHook
from novigi_common.hooks.datalake_s3_hook import DataLakeS3Hook

class S3toRedshiftStagingOperator(BaseOperator):
    """

    """

    @apply_defaults
    def __init__(
        self,
        *,
        schema: str,
        table: str,
        report_name: str,
        datalake_s3_conn_id: str = 'datalake_s3_connection_default',
        redshift_conn_id: str = 'redshift_postgres_default',
        verify: Optional[Union[bool, str]] = None,
        copy_options: Optional[List] = None,
        autocommit: bool = False,
        truncate_table: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.schema = schema
        self.table = table
        self.report_name = report_name
        self.datalake_s3_conn_id = datalake_s3_conn_id
        self.redshift_conn_id = redshift_conn_id
        self.verify = verify
        self.copy_options = copy_options or []
        self.autocommit = autocommit
        self.truncate_table = truncate_table

    def execute(self, context) -> None:
        postgres_hook = PostgresHook(postgres_conn_id=self.redshift_conn_id)
        datalake_s3_hook = DataLakeS3Hook(datalake_s3_conn_id=self.datalake_s3_conn_id)
        s3_bucket = datalake_s3_hook.get_s3_bucket_name()
        iam_role = datalake_s3_hook.get_s3_iam_role()
        current_execution_date = context['execution_date'].strftime("%Y/%m/%d")
        s3_key = self.report_name + '/' + current_execution_date + '/' + self.report_name + '.csv'
        #s3_hook = S3Hook(aws_conn_id=self.aws_conn_id, verify=self.verify)
        #credentials = s3_hook.get_credentials()
        copy_options = '\n\t\t\t'.join(self.copy_options)

        copy_statement = f"""
            COPY {self.schema}.{self.table}
            FROM 's3://{s3_bucket}/{s3_key}'
            iam_role '{iam_role}'
            {copy_options};
        """

        if self.truncate_table:
            truncate_statement = f'TRUNCATE TABLE {self.schema}.{self.table};'
            sql = f"""
            BEGIN;
            {truncate_statement}
            {copy_statement}
            COMMIT
            """
        else:
            sql = copy_statement

        self.log.info('Executing COPY command...')
        postgres_hook.run(sql, self.autocommit)
        self.log.info("COPY command complete...")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
from contextlib import closing
from typing import Iterable, Optional, Tuple, Union

import psycopg2
import psycopg2.extensions
import psycopg2.extras
from psycopg2.extensions import connection
from psycopg2.extras import DictCursor, NamedTupleCursor, RealDictCursor

from airflow.hooks.dbapi_hook import DbApiHook
from airflow.models.connection import Connection

CursorType = Union[DictCursor, RealDictCursor, NamedTupleCursor]


class PostgresHook(DbApiHook):
    """
    Interact with Postgres.

    You can specify ssl parameters in the extra field of your connection
    as ``{"sslmode": "require", "sslcert": "/path/to/cert.pem", etc}``.
    Also you can choose cursor as ``{"cursor": "dictcursor"}``. Refer to the
    psycopg2.extras for more details.

    Note: For Redshift, use keepalives_idle in the extra connection parameters
    and set it to less than 300 seconds.

    Note: For AWS IAM authentication, use iam in the extra connection parameters
    and set it to true. Leave the password field empty. This will use the
    "aws_default" connection to get the temporary token unless you override
    in extras.
    extras example: ``{"iam":true, "aws_conn_id":"my_aws_conn"}``
    For Redshift, also use redshift in the extra connection parameters and
    set it to true. The cluster-identifier is extracted from the beginning of
    the host field, so is optional. It can however be overridden in the extra field.
    extras example: ``{"iam":true, "redshift":true, "cluster-identifier": "my_cluster_id"}``
    """

    conn_name_attr = 'postgres_conn_id'
    default_conn_name = 'postgres_default'
    supports_autocommit = True

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.schema: Optional[str] = kwargs.pop("schema", None)
        self.connection: Optional[Connection] = kwargs.pop("connection", None)
        self.conn: connection = None

    def _get_cursor(self, raw_cursor: str) -> CursorType:
        _cursor = raw_cursor.lower()
        if _cursor == 'dictcursor':
            return psycopg2.extras.DictCursor
        if _cursor == 'realdictcursor':
            return psycopg2.extras.RealDictCursor
        if _cursor == 'namedtuplecursor':
            return psycopg2.extras.NamedTupleCursor
        raise ValueError(f'Invalid cursor passed {_cursor}')

    def get_conn(self) -> connection:
        """Establishes a connection to a postgres database."""
        conn_id = getattr(self, self.conn_name_attr)
        conn = self.connection or self.get_connection(conn_id)

        # check for authentication via AWS IAM
        if conn.extra_dejson.get('iam', False):
            conn.login, conn.password, conn.port = self.get_iam_token(conn)

        conn_args = dict(
            host=conn.host,
            user=conn.login,
            password=conn.password,
            dbname=self.schema or conn.schema,
            port=conn.port,
        )
        raw_cursor = conn.extra_dejson.get('cursor', False)
        if raw_cursor:
            conn_args['cursor_factory'] = self._get_cursor(raw_cursor)

        for arg_name, arg_val in conn.extra_dejson.items():
            if arg_name not in [
                'iam',
                'redshift',
                'cursor',
            ]:
                conn_args[arg_name] = arg_val

        self.conn = psycopg2.connect(**conn_args)
        return self.conn

    def copy_expert(self, sql: str, filename: str) -> None:
        """
        Executes SQL using psycopg2 copy_expert method.
        Necessary to execute COPY command without access to a superuser.

        Note: if this method is called with a "COPY FROM" statement and
        the specified input file does not exist, it creates an empty
        file and no data is loaded, but the operation succeeds.
        So if users want to be aware when the input file does not exist,
        they have to check its existence by themselves.
        """
        if not os.path.isfile(filename):
            with open(filename, 'w'):
                pass

        with open(filename, 'r+') as file:
            with closing(self.get_conn()) as conn:
                with closing(conn.cursor()) as cur:
                    cur.copy_expert(sql, file)
                    file.truncate(file.tell())
                    conn.commit()

    def bulk_load(self, table: str, tmp_file: str) -> None:
        """Loads a tab-delimited file into a database table"""
        self.copy_expert(f"COPY {table} FROM STDIN", tmp_file)

    def bulk_dump(self, table: str, tmp_file: str) -> None:
        """Dumps a database table into a tab-delimited file"""
        self.copy_expert(f"COPY {table} TO STDOUT", tmp_file)

    # pylint: disable=signature-differs
    @staticmethod
    def _serialize_cell(cell: object, conn: Optional[connection] = None) -> object:
        """
        Postgresql will adapt all arguments to the execute() method internally,
        hence we return cell without any conversion.

        See http://initd.org/psycopg/docs/advanced.html#adapting-new-types for
        more information.

        :param cell: The cell to insert into the table
        :type cell: object
        :param conn: The database connection
        :type conn: connection object
        :return: The cell
        :rtype: object
        """
        return cell

    def get_iam_token(self, conn: Connection) -> Tuple[str, str, int]:
        """
        Uses AWSHook to retrieve a temporary password to connect to Postgres
        or Redshift. Port is required. If none is provided, default is used for
        each service
        """
        from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook

        redshift = conn.extra_dejson.get('redshift', False)
        aws_conn_id = conn.extra_dejson.get('aws_conn_id', 'aws_default')
        aws_hook = AwsBaseHook(aws_conn_id, client_type='rds')
        login = conn.login
        if conn.port is None:
            port = 5439 if redshift else 5432
        else:
            port = conn.port
        if redshift:
            # Pull the custer-identifier from the beginning of the Redshift URL
            # ex. my-cluster.ccdre4hpd39h.us-east-1.redshift.amazonaws.com returns my-cluster
            cluster_identifier = conn.extra_dejson.get('cluster-identifier', conn.host.split('.')[0])
            client = aws_hook.get_client_type('redshift')
            cluster_creds = client.get_cluster_credentials(
                DbUser=conn.login,
                DbName=self.schema or conn.schema,
                ClusterIdentifier=cluster_identifier,
                AutoCreate=False,
            )
            token = cluster_creds['DbPassword']
            login = cluster_creds['DbUser']
        else:
            token = aws_hook.conn.generate_db_auth_token(conn.host, port, conn.login)
        return login, token, port

    @staticmethod
    def _generate_insert_sql(
        table: str, values: Tuple[str, ...], target_fields: Iterable[str], replace: bool, **kwargs
    ) -> str:
        """
        Static helper method that generate the INSERT SQL statement.
        The REPLACE variant is specific to MySQL syntax.

        :param table: Name of the target table
        :type table: str
        :param values: The row to insert into the table
        :type values: tuple of cell values
        :param target_fields: The names of the columns to fill in the table
        :type target_fields: iterable of strings
        :param replace: Whether to replace instead of insert
        :type replace: bool
        :param replace_index: the column or list of column names to act as
            index for the ON CONFLICT clause
        :type replace_index: str or list
        :return: The generated INSERT or REPLACE SQL statement
        :rtype: str
        """
        placeholders = [
            "%s",
        ] * len(values)
        replace_index = kwargs.get("replace_index")

        if target_fields:
            target_fields_fragment = ", ".join(target_fields)
            target_fields_fragment = f"({target_fields_fragment})"
        else:
            target_fields_fragment = ''

        sql = "INSERT INTO {} {} VALUES ({})".format(table, target_fields_fragment, ",".join(placeholders))

        if replace:
            if target_fields is None:
                raise ValueError("PostgreSQL ON CONFLICT upsert syntax requires column names")
            if replace_index is None:
                raise ValueError("PostgreSQL ON CONFLICT upsert syntax requires an unique index")
            if isinstance(replace_index, str):
                replace_index = [replace_index]
            replace_index_set = set(replace_index)

            replace_target = [
                "{0} = excluded.{0}".format(col) for col in target_fields if col not in replace_index_set
            ]
            sql += " ON CONFLICT ({}) DO UPDATE SET {}".format(
                ", ".join(replace_index),
                ", ".join(replace_target),
            )
        return sql



import json
import csv
import requests
import re
from http.client import responses
from jsonpath_ng import jsonpath, parse as jparse

from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.exceptions import AirflowException


class NovigiJsonToCSVExportOperator(BaseOperator):

    template_fields = ['api_headers']


    def __init__(self, api_url: str, req_type = "GET", output_csv = "", json_path: str = None, api_headers: dict = None, pay_load: dict = None, mapping: dict = {}, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.api_url = api_url
            self.req_type = req_type
            self.output_csv = output_csv
            self.json_path = json_path
            self.api_headers = api_headers
            self.pload = pay_load
            self.mapping = mapping


    def execute(self, context):

            response = requests.request(self.req_type, self.api_url ,headers = self.api_headers,data = self.pload)

            sucess_range = re.search(r'2[0-9][0-9]', str(response.status_code))

            if not (sucess_range):
                raise AirflowException("Response status code is " + str(response.status_code) + " and response status is '"+ responses[response.status_code]+"'")

            data = response.json()

            jsonpath_expression = jparse(self.json_path)

            match = jsonpath_expression.find(data)

            if len(match) > 0:

                result = match[0].value

            else:
                raise AirflowException('can not get match')

            # now we will open a file for writing
            data_file = open(self.output_csv,'w')

            # create the csv writer object
            csv_writer = csv.writer(data_file)

            # Writing headers of CSV file
            header = self.mapping.keys()

            csv_writer.writerow(header)

            for values_in_result in result:

                row = []

                for each_value in self.mapping.values():

                    jsonpath_exp = jparse(each_value)

                    matcher = jsonpath_exp.find(values_in_result)

                    mapping_value = ''

                    if len(matcher) > 0:
                        mapping_value = matcher[0].value

                    row.append(mapping_value)

                csv_writer.writerow(row)

            data_file.close()

            return True



from typing import List, Optional, Union

from airflow.models.baseoperator import BaseOperator

from airflow.utils.decorators import apply_defaults

from novigi_common.hooks.odbc_hook import OdbcHook
from novigi_common.hooks.datalake_s3_hook import DataLakeS3Hook

class MssqlToS3GenericOperator(BaseOperator):
    """
    This operator is to handle generic sql queries
    """

    template_fields = ("query",)

    @apply_defaults
    def __init__(
        self,
        *,
        query: str,
        report_name: str,
        datalake_s3_conn_id: str = 'datalake_s3_connection_default',
        odbc_conn_id: str = 'mssql_default',
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.query = query
        self.report_name = report_name
        self.datalake_s3_conn_id = datalake_s3_conn_id
        self.odbc_conn_id = odbc_conn_id


    def execute(self, context) -> None:
        odbc_hook = OdbcHook(odbc_conn_id=self.odbc_conn_id)
        datalake_s3_hook = DataLakeS3Hook(datalake_s3_conn_id=self.datalake_s3_conn_id)
        s3_bucket = datalake_s3_hook.get_s3_bucket_name()
        #iam_role = datalake_s3_hook.get_s3_iam_role()
        current_execution_date = context['execution_date'].strftime("%Y/%m/%d")
        s3_key = self.report_name + '/' + current_execution_date + '/' + self.report_name + '.csv'

        sql_query = self.query

        self.log.info('Prepared query: ' + sql_query)
        self.log.info('S3 key: ' + s3_key)

        data_frame = odbc_hook.get_pandas_df(sql_query)

        self.log.info("Fetch query complete...")

        s3_path = f's3://{s3_bucket}/{s3_key}'

        self.log.info("S3 Path " + s3_path)

        data_frame.to_csv(s3_path, index = False)

        self.log.info("S3 CSV creation complete...")



# pylint: disable=c-extension-no-member
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains ODBC hook."""
from typing import Any, Optional
from urllib.parse import quote_plus

import pyodbc

from airflow.hooks.dbapi_hook import DbApiHook
#from airflow.utils.helpers import merge_dicts


class OdbcHook(DbApiHook):
    """
    Interact with odbc data sources using pyodbc.

    See :ref:`howto/connection/odbc` for full documentation.
    """

    DEFAULT_SQLALCHEMY_SCHEME = 'mssql+pyodbc'
    conn_name_attr = 'odbc_conn_id'
    default_conn_name = 'odbc_default'
    supports_autocommit = True

    def __init__(
        self,
        *args,
        database: Optional[str] = None,
        driver: Optional[str] = None,
        dsn: Optional[str] = None,
        connect_kwargs: Optional[dict] = None,
        sqlalchemy_scheme: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        :param args: passed to DbApiHook
        :param database: database to use -- overrides connection ``schema``
        :param driver: name of driver or path to driver. overrides driver supplied in connection ``extra``
        :param dsn: name of DSN to use.  overrides DSN supplied in connection ``extra``
        :param connect_kwargs: keyword arguments passed to ``pyodbc.connect``
        :param sqlalchemy_scheme: Scheme sqlalchemy connection.  Default is ``mssql+pyodbc`` Only used for
          ``get_sqlalchemy_engine`` and ``get_sqlalchemy_connection`` methods.
        :param kwargs: passed to DbApiHook
        """
        super().__init__(*args, **kwargs)
        self._database = database
        self._driver = driver
        self._dsn = dsn
        self._conn_str = None
        self._sqlalchemy_scheme = sqlalchemy_scheme
        self._connection = None
        self._connect_kwargs = connect_kwargs

    @property
    def connection(self):
        """``airflow.Connection`` object with connection id ``odbc_conn_id``"""
        if not self._connection:
            self._connection = self.get_connection(getattr(self, self.conn_name_attr))
        return self._connection

    @property
    def database(self) -> Optional[str]:
        """Database provided in init if exists; otherwise, ``schema`` from ``Connection`` object."""
        return self._database or self.connection.schema

    @property
    def sqlalchemy_scheme(self) -> Optional[str]:
        """Database provided in init if exists; otherwise, ``schema`` from ``Connection`` object."""
        return (
            self._sqlalchemy_scheme
            or self.connection_extra_lower.get('sqlalchemy_scheme')
            or self.DEFAULT_SQLALCHEMY_SCHEME
        )

    @property
    def connection_extra_lower(self) -> dict:
        """
        ``connection.extra_dejson`` but where keys are converted to lower case.

        This is used internally for case-insensitive access of odbc params.
        """
        return {k.lower(): v for k, v in self.connection.extra_dejson.items()}

    @property
    def driver(self) -> Optional[str]:
        """Driver from init param if given; else try to find one in connection extra."""
        if not self._driver:
            driver = self.connection_extra_lower.get('driver')
            if driver:
                self._driver = driver
        return self._driver and self._driver.strip().lstrip('{').rstrip('}').strip()

    @property
    def dsn(self) -> Optional[str]:
        """DSN from init param if given; else try to find one in connection extra."""
        if not self._dsn:
            dsn = self.connection_extra_lower.get('dsn')
            if dsn:
                self._dsn = dsn.strip()
        return self._dsn

    @property
    def odbc_connection_string(self):
        """
        ODBC connection string
        We build connection string instead of using ``pyodbc.connect`` params because, for example, there is
        no param representing ``ApplicationIntent=ReadOnly``.  Any key-value pairs provided in
        ``Connection.extra`` will be added to the connection string.
        """
        if not self._conn_str:
            conn_str = ''
            if self.driver:
                conn_str += f"DRIVER={{{self.driver}}};"
            if self.dsn:
                conn_str += f"DSN={self.dsn};"
            if self.connection.host:
                conn_str += f"SERVER={self.connection.host};"
            database = self.database or self.connection.schema
            if database:
                conn_str += f"DATABASE={database};"
            if self.connection.login:
                conn_str += f"UID={self.connection.login};"
            if self.connection.password:
                conn_str += f"PWD={self.connection.password};"
            if self.connection.port:
                f"PORT={self.connection.port};"

            extra_exclude = {'driver', 'dsn', 'connect_kwargs', 'sqlalchemy_scheme'}
            extra_params = {
                k: v for k, v in self.connection.extra_dejson.items() if not k.lower() in extra_exclude
            }
            for k, v in extra_params.items():
                conn_str += f"{k}={v};"

            self._conn_str = conn_str
        return self._conn_str

    def merge_dicts(self, dict1, dict2):
        """
        Merge two dicts recursively, returning new dict (input dict is not mutated).

        Lists are not concatenated. Items in dict2 overwrite those also found in dict1.
        """
        merged = dict1.copy()
        for k, v in dict2.items():
            if k in merged and isinstance(v, dict):
                merged[k] = merge_dicts(merged.get(k, {}), v)
            else:
                merged[k] = v
        return merged


    @property
    def connect_kwargs(self) -> dict:
        """
        Returns effective kwargs to be passed to ``pyodbc.connect`` after merging between conn extra,
        ``connect_kwargs`` and hook init.

        Hook ``connect_kwargs`` precedes ``connect_kwargs`` from conn extra.

        String values for 'true' and 'false' are converted to bool type.

        If ``attrs_before`` provided, keys and values are converted to int, as required by pyodbc.
        """

        def clean_bool(val):  # pylint: disable=inconsistent-return-statements
            if hasattr(val, 'lower'):
                if val.lower() == 'true':
                    return True
                elif val.lower() == 'false':
                    return False
            else:
                return val

        conn_connect_kwargs = self.connection_extra_lower.get('connect_kwargs', {})
        hook_connect_kwargs = self._connect_kwargs or {}
        merged_connect_kwargs = self.merge_dicts(conn_connect_kwargs, hook_connect_kwargs)

        if 'attrs_before' in merged_connect_kwargs:
            merged_connect_kwargs['attrs_before'] = {
                int(k): int(v) for k, v in merged_connect_kwargs['attrs_before'].items()
            }

        return {k: clean_bool(v) for k, v in merged_connect_kwargs.items()}

    def get_conn(self) -> pyodbc.Connection:
        """Returns a pyodbc connection object."""
        conn = pyodbc.connect(self.odbc_connection_string, **self.connect_kwargs)
        return conn

    def get_uri(self) -> str:
        """URI invoked in :py:meth:`~airflow.hooks.dbapi_hook.DbApiHook.get_sqlalchemy_engine` method"""
        quoted_conn_str = quote_plus(self.odbc_connection_string)
        uri = f"{self.sqlalchemy_scheme}:///?odbc_connect={quoted_conn_str}"
        return uri

    def get_sqlalchemy_connection(
        self, connect_kwargs: Optional[dict] = None, engine_kwargs: Optional[dict] = None
    ) -> Any:
        """Sqlalchemy connection object"""
        engine = self.get_sqlalchemy_engine(engine_kwargs=engine_kwargs)
        cnx = engine.connect(**(connect_kwargs or {}))
        return cnx



from airflow.plugins_manager import AirflowPlugin
from novigi_common.operators.custom_file_load_operator import CustomFileProcessingOperator
from novigi_common.hooks.sample_custom_hook import CustomHook


# Defining the plugin class
class CustomPlugin(AirflowPlugin):
    name = "custom_plugin"
    operators = [CustomFileProcessingOperator]
    hooks = [CustomHook]


from airflow.hooks.dbapi_hook import DbApiHook

'''
TODO: Further improve this Hook

Conn Id: datalake_s3_connection_default
Conn Type: S3
Extra: {"aws_s3_bucket_name":"_your_aws_bucket_name_", "aws_s3_iam_role": "_your_aws_s3_iam_role_"}
Leave all the other fields (Host, Schema, Login) blank.

'''

class DataLakeS3Hook(DbApiHook):

    conn_name_attr = 'datalake_s3_conn_id'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        conn_id = getattr(self, self.conn_name_attr)
        self.connection = self.get_connection(conn_id)

    def get_s3_bucket_name(self):
        return self.connection.extra_dejson['aws_s3_bucket_name']

    def get_s3_iam_role(self):
        return self.connection.extra_dejson['aws_s3_iam_role']





from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='novigi_common',
    version="1.0.2",
    description='Novigi Custom Airflow operators, hooks and plugins',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT', 
    packages=['novigi_common','novigi_common.operators', 'novigi_common.hooks', 'novigi_common.custom_plugings'],
    install_requires=['requests','jsonpath_ng','pandas', 'pyodbc', 'psycopg2'],
    setup_requires=['setuptools', 'wheel'],
    author='Novigi',
    author_email='integration@novigi.com.au',
)



from typing import List, Optional, Union

from airflow.models.baseoperator import BaseOperator

from airflow.utils.decorators import apply_defaults

from novigi_common.hooks.odbc_hook import OdbcHook
from novigi_common.hooks.datalake_s3_hook import DataLakeS3Hook

class MssqlToS3Operator(BaseOperator):
    """

    """

    @apply_defaults
    def __init__(
        self,
        *,
        schema: str = 'dbo',
        table: str,
        report_name: str,
        datalake_s3_conn_id: str = 'datalake_s3_connection_default',
        odbc_conn_id: str = 'mssql_default',
        field_list: Optional[List] = None,
        last_update_field: str = 'last_update',
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.schema = schema
        self.table = table
        self.report_name = report_name
        self.datalake_s3_conn_id = datalake_s3_conn_id
        self.odbc_conn_id = odbc_conn_id
        self.field_list = field_list or []
        self.last_update_field = last_update_field


    def execute(self, context) -> None:
        odbc_hook = OdbcHook(odbc_conn_id=self.odbc_conn_id)
        datalake_s3_hook = DataLakeS3Hook(datalake_s3_conn_id=self.datalake_s3_conn_id)
        s3_bucket = datalake_s3_hook.get_s3_bucket_name()
        #iam_role = datalake_s3_hook.get_s3_iam_role()
        current_execution_date = context['execution_date'].strftime("%Y/%m/%d")
        s3_key = self.report_name + '/' + current_execution_date + '/' + self.report_name + '.csv'

        query_start_date = context['execution_date'].strftime("%Y-%m-%d")
        query_end_date = context['execution_date'].strftime("%Y-%m-%d")

        if context['dag_run'].conf is not None:
            query_start_date = context['dag_run'].conf.get('start_date', None) or context['execution_date'].strftime("%Y-%m-%d")
            query_end_date = context['dag_run'].conf.get('end_date', None) or context['execution_date'].strftime("%Y-%m-%d")

        filter_string = '*'
        if len(self.field_list) > 0:
            self.log.info("Filtering the columns "+ str(self.field_list))
            filter_string = ','.join(self.field_list)

        sql_query = f'SELECT {filter_string} FROM [{self.schema}].[{self.table}] WHERE {self.last_update_field} BETWEEN \'{query_start_date} 00:00:00\' AND \'{query_end_date} 23:59:59\''

        self.log.info('Prepared query: ' + sql_query)
        self.log.info('S3 key: ' + s3_key)

        data_frame = odbc_hook.get_pandas_df(sql_query)

        self.log.info("Fetch query complete...")

        s3_path = f's3://{s3_bucket}/{s3_key}'

        self.log.info("S3 Path " + s3_path)

        data_frame.to_csv(s3_path, index = False)

        self.log.info("S3 CSV creation complete...")
